from csv import DictWriter

from nerium.query import serialize_stream
from nerium.streaming import BufferWriter


class CsvStreamWriter(BufferWriter):
    """Serializes dicts to a StringIO buffer"""

    def __init__(self, stream, first_record):
        self.stream = stream
        self.writer = DictWriter(stream, fieldnames=first_record.keys())
        self.writer.writeheader()

    def write(self, record):
        """Absraction to support writers with methods other than writerow"""

        self.writer.writerow(record)


def results_to_csv(query_name, **kwargs):
    """Generate CSV from result data"""

    return serialize_stream(query_name, CsvStreamWriter, **kwargs)



import importlib.util
import os
from importlib import import_module


def get_format(format_="default"):
    """ Find format schema in $FORMAT_PATH or nerium/schema
        NOTE: if `format_` is not found, this falls back to "default"
    """
    # Load format from builtin schema if exists
    try:
        format_mod = import_module(f"nerium.schema.{format_}")
    except ModuleNotFoundError:
        # Attempt to load custom format from $FORMAT_PATH
        try:
            format_path = os.getenv("FORMAT_PATH", "format_files")
            spec = importlib.util.spec_from_file_location(
                "format_mod", f"{format_path}/{format_}.py"
            )
            format_mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(format_mod)
        # Fallback to default schema if no other match
        except FileNotFoundError:
            format_mod = import_module("nerium.schema.default")
    schema = format_mod.ResultSchema()
    return schema


# TODO: dump from format schema here(?)



from pathlib import Path

import requests
from git import Repo
from git.exc import InvalidGitRepositoryError

from nerium import __version__


def get_local_head_commit():
    workdir = Path(__file__).resolve().parents[1]
    repo = Repo(workdir)
    commit = repo.head.commit.hexsha
    return commit


def get_commit_for_version(version=__version__):
    v = f"v{version}"

    try:
        resp = requests.get("https://api.github.com/repos/OAODEV/nerium/tags")
        tags = resp.json()
        vtag = list(filter(lambda t: v == t["name"], tags))
        vtag = vtag[0]
        commit_url = vtag["commit"]["url"]
        resp = requests.get(commit_url)
        commit = resp.json()["url"]
    except (
        IndexError,
        KeyError,
        TypeError,
        requests.exceptions.ConnectionError,
        ValueError,
    ):
        try:
            commit = get_local_head_commit()
        except InvalidGitRepositoryError:
            commit = "Unable to retrieve commit here"
    return commit


if __name__ == "__main__":
    commit = get_commit_for_version()
    print(commit)



import re
from pathlib import Path
from types import SimpleNamespace

import sqlparse
from raw import db
from sqlparse.sql import Identifier, IdentifierList

from nerium.query import parse_query_file


def list_reports():
    """Return list of available report names from query dir"""
    flat_queries = db.list_queries()
    query_names = [Path(i).stem for i in flat_queries]

    query_names.sort()
    reports = dict(reports=query_names)
    return reports


def columns_from_metadata(query):
    """Return `columns` block from query front matter, if present"""
    columns = None
    if "columns" in query.metadata.keys():
        columns = query.metadata.pop("columns")
    return columns


def columns_from_statement(query):
    """Parse columns from SELECT list"""
    # TODO: can we base this on SQLA attributes instead of using sqlparse?
    columns = []
    parsed_query = sqlparse.parse(query.statement)[0]
    for tkn in parsed_query.tokens:
        if isinstance(tkn, IdentifierList):
            for id_ in tkn:
                if isinstance(id_, Identifier):
                    columns.append(id_.get_name())
    return columns


def params_from_metadata(query):
    params = None
    if "params" in query.metadata.keys():
        params = query.metadata.pop("params")
    return params


def params_from_statement(query):
    # TODO: can we base this on SQLA attributes instead of using sqlparse?
    param_regex = re.compile("(?<!\\w)\\:\\w+")
    param_list = [i.strip(":") for i in re.findall(param_regex, query.statement)]
    return param_list


def describe_report(query_name):
    report_query = parse_query_file(query_name)
    if report_query.error:
        return report_query
    params = params_from_metadata(report_query) or params_from_statement(report_query)
    columns = columns_from_metadata(report_query) or columns_from_statement(
        report_query
    )
    report_description = SimpleNamespace(
        error=report_query.error,
        name=report_query.name,
        columns=columns,
        params=params,
        metadata=report_query.metadata,
    )
    return report_description



import abc

from io import StringIO

BUFFER_SIZE = 16384


def initialize_stream(iterable, writer_constructor):
    """Initialize a stream and writer. Popping the first record from
    the iterable allows both for error handling prior to entering a
    stream generator as well as header formatting. The first record will
    be written to the `stream` buffer here.

    Args:
        `iterable`: an iterable of results, e.g. a generator streaming
            a sqlalchemy cursor
        `writer_constructor`: a class or callable whose return
            implements the BufferWriter interface

    Returns:
        A tuple containing, respectively, a stream (StringIO stream
        buffer) and a writer (implementing BufferWriter). These can be
        passed to yield_stream along with the original `iterable` to
        yield the stream until completion

    """

    stream = StringIO()

    try:
        first = next(iterable)
    except StopIteration:
        return stream, None

    writer = writer_constructor(stream, first)
    writer.write(first)

    return stream, writer


def yield_stream(iterable, stream, writer, **kwargs):
    """Buffers contents of iterable to stream and yields blocks of
    BUFFER_SIZE until completion. Arguments `stream` and `writer` are
    expected to be the return values of `initialize_stream`.

    Args:
        `iterable`: an iterable of results, e.g. a generator streaming
            a sqlalchemy cursor
        `stream`: an iterable of results, e.g. a generator streaming
            a sqlalchemy cursor
        `writer`: serializes each record in the iterable to the
            StringIO buffer. Implements the BufferWriter interface.

    Yields:
        bytes: serialized records form iterable

    """

    for row in iterable:
        writer.write(row)
        if stream.tell() > BUFFER_SIZE:
            yield _flush(stream)
    yield _flush(stream)


def _flush(stream):
    """Consume the current StringIO buffer and return the contents"""

    stream.seek(0)
    data = stream.read()
    stream.truncate(0)
    stream.seek(0)

    return data


class BufferWriter(metaclass=abc.ABCMeta):
    """Interface specifying serialization to a StringIO buffer"""

    @abc.abstractmethod
    def __init__(self, stream, first_record):
        pass

    @abc.abstractmethod
    def write(self, record):
        pass



import re
from collections import namedtuple
from datetime import datetime

import s3fs
import yaml
from raw import db
from nerium import streaming


def init_query(query_name):
    """Initialize query namedtuple from file name"""
    Query = namedtuple(
        "Query",
        [
            "name",
            "executed",
            "error",
            "metadata",
            "path",
            "statement",
            "result",
            "status_code",
        ],
    )
    # `nametuple.defaults` option only available in >=3.7
    # .__new__.defaults__ works with older versions
    Query.__new__.__defaults__ = (False, {}, "", "", "", 200)
    query_obj = Query(
        name=query_name,
        executed=datetime.utcnow().isoformat(),
        path=db.path_by_name(query_name),
    )
    return query_obj


def extract_metadata(query_string):
    """Find frontmatter comment in query string and load yaml from it if present"""
    metadata = {}
    meta_comment = re.search(r"---[\s\S]+?---", query_string, re.MULTILINE)
    if meta_comment:
        meta_string = meta_comment[0].strip("---")
        metadata = yaml.safe_load(meta_string)

    return metadata


def read_query_file(path):
    if path.startswith("s3://"):
        # Read file from S3 and convert to string with decode()
        fs = s3fs.S3FileSystem(anon=False)
        with fs.open(path, "rb") as f:
            statement = f.read().decode()
    else:
        # Read file as text from local filesystem
        with open(path, "rt") as f:
            statement = f.read()
    return statement


def parse_query_file(query_name):
    """Parse query file and add query statement and metadata."""
    query_obj = init_query(query_name)

    if not query_obj.path:
        query_obj = query_obj._replace(
            error=f"No query found matching '{query_name}'", status_code=404
        )
        return query_obj

    statement = read_query_file(query_obj.path)
    metadata = extract_metadata(statement)
    query_obj = query_obj._replace(metadata=metadata, statement=statement)
    return query_obj


def query_decorator(func):
    """Decorates a "result" query method, e.g. a sqla-raw method, in a
    Query for parsing and execution error handling.

    """

    def inner(query_name, *args, **kwargs):
        query = parse_query_file(query_name)

        # Bail if parser captures error:
        if query.error:
            return query

        try:
            result = func(query_name, *args, **kwargs)
            query = query._replace(result=result)
        except Exception as e:
            query = query._replace(error=repr(e), status_code=400)

        return query
    return inner


@query_decorator
def get_result_set(query_name, **kwargs):
    """Call parse_query_file, then submit query and return new object with result"""

    return db.result_by_name(query_name, **kwargs)


@query_decorator
def serialize_stream(query_name, writer_constructor, **kwargs):
    """Call parse_query_file, then submit query and return generator with result"""

    result = db.stream_result_by_name(query_name, **kwargs)
    # Initialization here (outside of a generator) allows for
    # exception handling in `query_decorator` and an HTTP error
    # return prior to starting a 200 streaming HTTP
    # response. Otherwise, an exception raised from within the
    # generator iteration would occur after the stream response is returned
    stream, writer = streaming.initialize_stream(result, writer_constructor)

    return streaming.yield_stream(result, stream, writer, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import functools
import os

from flask import request


def require_api_key(func):
    """Decorator that requires a request to have a valid X-API-Key header
    if API_KEY is set in the environment
    """

    @functools.wraps(func)
    def decorator(*args, **kwargs):
        API_KEY = os.getenv("API_KEY")
        # If API_KEY is not provided in environment, just pass through
        if not API_KEY:
            return func(*args, **kwargs)

        # Otherwise API_KEY is set. Require valid key sent in header:
        api_key = request.headers.get("X-API-Key")
        # 400 error if required header is not set
        if not api_key:
            return {"error": "Missing API key"}, 400
        # Check if API key is correct and valid, 403 if not
        elif api_key == API_KEY:
            return func(*args, **kwargs)
        else:
            return {"error": "The provided API key is not valid"}, 403

    return decorator
